/************************************************************************/
/*                                                                      */
/* Copyright (C) 2004 Mattias Eklf <tias@ludd.ltu.se>                  */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 2, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            */
/* 02111-1307, USA.                                                     */
/*                                                                      */
/************************************************************************/

#include <licq/licq_constants.h>
#include <licq/licq_events.h>
#include <licq/licq_icqd.h>
#include <licq/licq_plugin.h>

#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config.h"

#define INFOFS_DIR ".licq/infofs"

#define STATUS_FILE_NAME "status"
#define PENDING_FILE_NAME "pending"
#define NUM_EVENTS_FILE_NAME "num_events"

static CICQDaemon *icq_daemon;

static int status_file;
static int pending_file;
static int num_events_file;

const char *LP_Name()
{
	static const char *name = "infofs";

	return (name);
}

const char *LP_Version()
{
	static const char *version = VERSION;

	return (version);
}

const char *LP_Status()
{
	static const char *running_str = "Running";

	return (running_str);
}

const char *LP_Description()
{
	static const char *description = "Creates a little 'filesystem', '.licq/infofs', which contains licq state info.";

	return (description);
}

const char *LP_Usage()
{
	static const char *usage = "";

	return (usage);
}

void write_status(unsigned short status, bool inv)
{
	const char *online_str = "On";
	const char *away_str = "Away";
	const char *na_str = "N/A";
	const char *dnd_str = "DND";
	const char *occ_str = "Occ";
	const char *ffc_str = "FFC";
	const char *offline_str = "Off";
	
	char tmp[100];
	const char *str;

	switch (status) {
	case ICQ_STATUS_ONLINE:
	        str = online_str;
		break;
	case ICQ_STATUS_AWAY:
		str = away_str;
		break;
	case ICQ_STATUS_NA:
		str = na_str;
		break;
	case ICQ_STATUS_DND:
		str = dnd_str;
		break;
	case ICQ_STATUS_OCCUPIED:
		str = occ_str;
		break;
	case ICQ_STATUS_FREEFORCHAT:
		str = ffc_str;
		break;
	case ICQ_STATUS_OFFLINE:
	default:
		str = offline_str;
		break;
	};
	
	if (inv)
		sprintf(tmp, "(%s)\n", str);
	else
		sprintf(tmp, "%s\n", str);

	flock(status_file, LOCK_EX);
	ftruncate(status_file, 0);
	lseek(status_file, 0, SEEK_SET);
	write(status_file, tmp, strlen(tmp));
	fsync(status_file);
	flock(status_file, LOCK_UN);
}

void write_pending(unsigned short sub_cmd)
{
	char tmp[100];
	
	switch (sub_cmd) {
	case ICQ_CMDxSUB_MSG:
		strcpy(tmp, "MSG");
		break;
	case ICQ_CMDxSUB_CHAT:
		strcpy(tmp, "CHAT");
		break;			
	case ICQ_CMDxSUB_FILE:
		strcpy(tmp, "FILE");
		break;
	case ICQ_CMDxSUB_URL:
		strcpy(tmp, "URL");
		break;
	case ICQ_CMDxSUB_SMS:
		strcpy(tmp, "SMS");
		break;
	default:
		// Every event we don't know what it is we call "SYSTEM".
		strcpy(tmp, "SYSTEM");
		break;	
	}
	
	strcat(tmp, "\n");

	flock(pending_file, LOCK_EX);
	ftruncate(pending_file, 0);
	lseek(pending_file, 0, SEEK_SET);
	write(pending_file, tmp, strlen(tmp));
	flock(pending_file, LOCK_UN);
}

void write_num_events(int num_events)
{
	char tmp[100];
	sprintf(tmp, "%i\n", num_events);

	flock(num_events_file, LOCK_EX);
	ftruncate(num_events_file, 0);
	lseek(num_events_file, 0, SEEK_SET);
	write(num_events_file, tmp, strlen(tmp));
	flock(num_events_file, LOCK_UN);
}

bool LP_Init(int argc, char **argv)
{
	char *infofs_abs_path;
	char *status_file_abs_path;
	char *pending_file_abs_path;
	char *num_events_file_abs_path;
	
	char *home = getenv("HOME");

	// Make infofs dir if it doesn't exist.
	infofs_abs_path = new char[strlen(home)+1+strlen(INFOFS_DIR)+1];
	strcpy(infofs_abs_path, home);
	strcat(infofs_abs_path, "/");
	strcat(infofs_abs_path, INFOFS_DIR);
	mkdir(infofs_abs_path, 0);

	status_file_abs_path = new char[strlen(infofs_abs_path)+1+strlen(STATUS_FILE_NAME)+1];
	strcpy(status_file_abs_path, infofs_abs_path);
	strcat(status_file_abs_path, "/");
	strcat(status_file_abs_path, STATUS_FILE_NAME);

	status_file = open(status_file_abs_path, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (status_file == -1) {
		delete infofs_abs_path;
		delete status_file_abs_path;
		return (false);
	}
	
	pending_file_abs_path = new char[strlen(infofs_abs_path)+1+strlen(PENDING_FILE_NAME)+1];
	strcpy(pending_file_abs_path, infofs_abs_path);
	strcat(pending_file_abs_path, "/");
	strcat(pending_file_abs_path, PENDING_FILE_NAME);

	pending_file = open(pending_file_abs_path, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (pending_file == -1) {
		delete infofs_abs_path;
		delete pending_file_abs_path;
		return (false);
	}
	
	num_events_file_abs_path = new char[strlen(infofs_abs_path)+1+strlen(NUM_EVENTS_FILE_NAME)+1];
	strcpy(num_events_file_abs_path, infofs_abs_path);
	strcat(num_events_file_abs_path, "/");
	strcat(num_events_file_abs_path, NUM_EVENTS_FILE_NAME);
	
	num_events_file = open(num_events_file_abs_path, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (num_events_file == -1) {
		delete infofs_abs_path;
		delete num_events_file_abs_path;
		return (false);
	}

	delete infofs_abs_path;
	delete status_file_abs_path;
	delete pending_file_abs_path;
	delete num_events_file_abs_path;

	// Get and write initial status. Assume we're offline, since
	// gUserManager.FetchUser() will break on startup. If Licq has just
	// been started, offline will be the right status.
	write_status(ICQ_STATUS_OFFLINE, false);
	write_num_events(0);

	return (true);
}

void process_signal(CICQSignal *s)
{
	switch (s->Signal()) {
	case SIGNAL_UPDATExUSER:
		if (s->SubSignal() == USER_EVENTS && s->Uin() != gUserManager.OwnerUin()) {
			ICQOwner *o = gUserManager.FetchOwner(LOCK_R);
			unsigned short num_owner_events = o->NewMessages();
			gUserManager.DropOwner();
			unsigned short num_user_events = ICQUser::getNumUserEvents() - num_owner_events;			

			if (s->Argument() > 0 && num_user_events == 1) {
				ICQUser *u = gUserManager.FetchUser(s->Uin(), LOCK_R);
				CUserEvent *e = u->EventPeekId(s->Argument());
				gUserManager.DropUser(u);

				write_pending(e->SubCommand()); 
			}

			write_num_events(num_user_events);
		} else if (s->SubSignal() == USER_STATUS) {
			ICQOwner *o = gUserManager.FetchOwner(LOCK_R);
			unsigned short status = o->Status();
			bool inv = o->StatusInvisible();
			gUserManager.DropOwner();

			write_status(status, inv);
		}
		break;
	default:
		break;
	}
	delete s;
}

int LP_Main(CICQDaemon *daemon)
{
	icq_daemon = daemon;
	int pipe = icq_daemon->RegisterPlugin(SIGNAL_UPDATExUSER);

	for (;;) {
		char buf[1];
		read(pipe, buf, 1);

		switch (buf[0]) {
		case 'X': // Shut down
			write_status(ICQ_STATUS_OFFLINE, false);
			write_num_events(0);
			
			close(status_file);
			close(pending_file);
			close(num_events_file);			
			return(0);
		case 'S': // Caught signal
			process_signal(icq_daemon->PopPluginSignal());
			break;
		default:
			break;
		}
	}

	// We shouldn't end up here.
	assert(true);
	return (0);
}
